-- view, backgrounds, tiles, sprites
require("util")

class("Rect")

function Rect:init(x,y,w,h)
    self.x = x
    self.y = y
    self.w = w
    self.h = h
    if self.w == nil or self.h == nil then
        self.w = self.x
        self.h = self.y
        self.x = 0
        self.y = 0
    end
end

function Rect:top(v) if v ~= nil then self.y = v end return self.y end
function Rect:right(v) if v ~= nil then self.x = v-self.w end return self.x+self.w end
function Rect:bottom(v) if v ~= nil then self.y = v-self.h end return self.y+self.h end
function Rect:left(v) if v ~= nil then self.x = v end return self.x end
function Rect:centerx(v) if v ~= nil then self.x = v-self.w/2 end return self.x+self.w/2 end
function Rect:centery(v) if v ~= nil then self.y = v-self.h/2 end return self.y+self.h/2 end
    
function Rect:clamp(r)
    local tmp
    self.x = max(self.x,r.x)
    self.y = max(self.y,r.y)
    tmp = min(self:right(),r:right())
    self.x = tmp-self.w
    tmp = min(self:bottom(),r:bottom())
    self.y = tmp-self.h
    tmp = self.w - r.w
    if tmp > 0 then self.x = -tmp/2 end
    tmp = self.h - r.h
    if tmp > 0 then self.y = -tmp/2 end
end

function Rect:collidepoint(p)
    return (self.x <= p.x and self.y <= p.y and (self.x+self.w) > p.x and (self.y+self.h) > p.y)
end

function Rect:colliderect(r)
    return (not (r:right() <= self:left() or self:right() <= r:left() or r:bottom() <= self:top() or self:bottom() <= r:top()))
end

function Rect:copy()
    return Rect(self.x,self.y,self.w,self.h)
end

class("Sprite")

function Sprite:init(image,rect,shape)
    self.image = image
    self.shape = shape or Rect(0,0,rect.w,rect.h)
    self.rect = rect
    self.angle = 0.0
    self.color = 0xffffff
    self.alpha = 1.0
    self.mode = 0
end

function Sprite:draw()
    cuz.image_draw(self.image,
        round(self.rect.x+self.shape.x),round(self.rect.y+self.shape.y),
        self.shape.w,self.shape.h,
        0,0,self.angle,
        self.color,self.alpha,self.mode)
end

class("Tile")

function Tile:init(image)
    self.image = image
end

class("Scene")

function Scene:init(bounds,view)
    self.items = {}
    self.bounds = bounds
    self.view = view
end

function Scene:draw()
    self.view.x = round(self.view.x)
    self.view.y = round(self.view.y)
    self.view:clamp(self.bounds)
    cuz.view_set(self.view.x,self.view.y,self.view.w,self.view.h)
    for n,obj in pairs(self.items) do
        obj:draw(self)
    end
    self:_letterbox()
end

function Scene:_letterbox()
    local z = 99999
    local x = self.view.x
    local y = self.view.y
    cuz.image_draw("black",x+-z,y+-z,z+self.view.w+z,z)
    cuz.image_draw("black",x+self.view.w,y+-z,z,z+self.view.h+z)
    cuz.image_draw("black",x+-z,y+self.view.h,z+self.view.w+z,z)
    cuz.image_draw("black",x+-z,y+-z,z,z+self.view.h+z)
end

function Scene:add(obj,name)
    obj.name = name
    self.items[#self.items+1] = obj
end

function Scene:find(name)
    for n,obj in pairs(self.items) do
        if obj.name == name then return obj end
    end
end

function Scene:loop(t)
    for n,obj in pairs(self.items) do
        if obj.loop then obj:loop(self,t) end
    end
end

function Scene:event(e)
    for n,obj in pairs(self.items) do
        if obj.event then obj:event(self,e) end
    end
end
        

class("Tilemap")

function Tilemap:init(tiles,data,tile_size)
    self.tiles = tiles
    self.data = data
    self.tile_size = tile_size
--     for y=1,rect.h do
--         local row = {}
--         self.data[y] = row
--         for x=1,rect.w do
--             row[x] = 0
--         end
--     end
end

function Tilemap:set(x,y,v)
    x = round(x)
    y = round(y)
    local w = #self.data[1]
    local h = #self.data
    if x < 1 or y < 1 or x > w or y > h then return end
    self.data[y][x] = v
end

function Tilemap:get(x,y)
    x = round(x)
    y = round(y)
    local w = #self.data[1]
    local h = #self.data
    if x < 1 or y < 1 or x > w or y > h then return 0 end
    return self.data[y][x]
end

function Tilemap:draw(scene)
    local tw = self.tile_size.w
    local th = self.tile_size.h
    local x_min = max(1,math.floor(scene.view.x / tw))
    local x_max = min(#self.data[1],math.ceil(scene.view:right() / tw))
    local y_min = max(1,math.floor(scene.view.y / th))
    local y_max = min(#self.data,math.ceil(scene.view:bottom() / th))
--     print("view: x="..x_min.."..."..x_max.." y="..y_min.."..."..y_max)
    for y=y_min,y_max do
        local row = self.data[y]
        for x=x_min,x_max do
            local n = row[x]
            if n ~= 0 then
                local tile = self.tiles[n]
                cuz.image_draw(tile.image,
                    x * tw,y * th,
                    tw,th)
            end
        end
    end
end
    

class("Background")

function Background:init(image,rect)
    self.image = image
    self.rect = rect
end

function Background:draw(scene)
    -- background size
    local w = self.rect.w
    local h = self.rect.h
    -- background
    local bw = min(scene.bounds.w,w)
    local bh = min(scene.bounds.h,h)
    -- background shift
    local sw = max(0,bw-scene.view.w)
    local sh = max(0,bh-scene.view.h)
    -- view corner
    local vx = scene.view.x - scene.bounds.x
    local vy = scene.view.y - scene.bounds.y
    -- view shift
    local zw = max(0,scene.bounds.w-scene.view.w)
    local zh = max(0,scene.bounds.h-scene.view.h)
    -- figure offsets
    local ox = (bw-w)/2
    local oy = (bh-h)/2
    -- figure x,y
    local x = scene.view.x - ((vx * sw) / max(1,zw)) + ox
    local y = scene.view.y - ((vy * sh) / max(1,zh)) + oy
    -- re-center for small backgrounds
    if w < scene.view.w then x = scene.view.x + (scene.view.w-w) / 2 end
    if h < scene.view.h then y = scene.view.y + (scene.view.h-h) / 2 end
    cuz.image_draw(self.image,
        round(x),round(y),
        w,h)
end

TEST = false
if TEST then
    r = Rect(1,2,3,4)
    print(r.x)
    print(r:top())
    print(r:bottom())
    print(r:left())
    print(r:right())
end
